import json
from os import path, makedirs


def get_abi(contract):
    abi_file_path = f"{path.dirname(__file__)}/abi/"
    with open(f'{abi_file_path}/{contract}.json') as json_file:
        return json.load(json_file)


def initialize_contract(w3, contract_address, contract_name):
    return w3.eth.contract(
        address=contract_address,
        abi=get_abi(contract_name)
    )


import sys
import logging
from web3 import Web3
from tbtc.settings import (
    LOG_STDOUT,
    LOG_FORMAT,
    LOG_LEVEL,
)
from tbtc.constants import contracts


logger = logging.getLogger(__name__)


def setup_logging():
    """
    Add logging format to logger used for debugging and info
    """
    handler = logging.StreamHandler(
        sys.stdout if LOG_STDOUT else sys.stderr)
    formatter = logging.Formatter(LOG_FORMAT)
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)
    logging.getLogger().setLevel(LOG_LEVEL)

    # removing logs from noisy libraries
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("web3").setLevel(logging.WARNING)


def init_web3(node_url):
    return Web3(Web3.HTTPProvider(node_url))


def get_contracts(version, network):
    return contracts[version][network]


def init_account(w3, private_key):
    return w3.eth.account.from_key(private_key)


from riemann.encoding.addresses import make_p2wpkh_address

def point_to_pubkey(x, y, compressed):
    if compressed:
        int_y = int.from_bytes(y, 'big')
        if int_y & 1:  # odd root
            return b'\x03' + x.rjust(32, b'\x00')
        else:           # even root
            return b'\x02' + x.rjust(32, b'\x00')
    else:
        return b'\x04' + x.rjust(32, b'\x00') + y.rjust(32, b'\x00')


def point_to_p2wpkh_address(x, y):
    return make_p2wpkh_address(point_to_pubkey(x, y, True))


import logging
from tbtc.session import (
    init_account,
    get_contracts,
)
from tbtc.utils import initialize_contract
from tbtc.deposit import Deposit
import riemann

logger = logging.getLogger(__name__)

class TBTC():

    def __init__(self, version, web3, btc_network, private_key):
        self.version = version
        self.w3 = web3
        if self.w3.eth.chainId == 1:
            self.eth_network = "mainnet"
        elif self.w3.eth.chainId == 3:
            self.eth_network = "ropsten"
        else:
            raise KeyError(f"Ethereum network {self.w3.eth.chainId} not defined")
        self.btc_network = btc_network
        if self.btc_network == 'testnet':
            riemann.select_network('bitcoin_test')
        self.account = init_account(self.w3, private_key)
        self.contracts = get_contracts(self.version, self.eth_network)
        self.system = initialize_contract(self.w3, self.contracts["TBTCSystem"], "TBTCSystem")
        self.deposit_factory = initialize_contract(self.w3, self.contracts["DepositFactory"], "DepositFactory")
        logger.info(f"Initialized tBTC with version {self.version} and eth network {self.eth_network}")
    

    def get_available_lot_sizes(self):
        return self.system.functions.getAllowedLotSizes().call()

    
    def create_deposit(self, lot_size):
        logger.info("Initiating deposit...")
        value = self.system.functions.getNewDepositFeeEstimate().call()
        create_deposit_call = self.deposit_factory.functions.createDeposit(lot_size)
        receipt = self._manage_transaction(create_deposit_call, gas_limit=1600000, value=value)
        logs = self.system.events.Created().processReceipt(receipt)
        deposit = Deposit(
            self, 
            logs[0]['args']['_depositContractAddress'],
            logs[0]['args']['_keepAddress']
            )
        return deposit


    def _manage_transaction(self, function_call, gas_limit, value=0):
        unsigned_txn = function_call.buildTransaction({
            'chainId': self.w3.eth.chainId,
            'nonce': self.w3.eth.getTransactionCount(self.account.address),
            'gas': gas_limit,
            'value': value
        })
        signed_txn = self.w3.eth.account.sign_transaction(
            unsigned_txn, self.account.key)
        tx_hash = self.w3.eth.sendRawTransaction(
            signed_txn.rawTransaction)
        logger.info(f"Sent the transaction {tx_hash.hex()}")
        logger.info("Waiting for transaction to be mined ...")
        return self.w3.eth.waitForTransactionReceipt(tx_hash)


    def _get_existing_public_key_event(self, deposit_address):
        event_filter = self.system.events.RegisteredPubkey.createFilter(
            fromBlock='earliest',
            toBlock='latest',
            argument_filters={'_depositContractAddress': deposit_address}
        )
        return event_filter.get_all_entries()



import os

HOME_DIR = os.path.expanduser("~")

# Logging settings
LOG_STDOUT = os.getenv("LOG_STDOUT", "TRUE")
LOG_FORMAT = os.getenv(
    "LOG_FORMAT", "[%(asctime)s][%(levelname)s][%(name)s] %(message)s")
LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")



import socket
import ssl
import json
import time
import errno

class Electrum():
    def __init__(self, host, port, protocol, timeout=5):
        self.host = host
        self.port = port
        self.protocol = protocol
        self.timeout = timeout
        self.connection = None
        self._connect()

    def _connect(self):
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection.settimeout(self.timeout)
        if self.protocol=="ssl":
            self.connection = ssl.wrap_socket(self.connection)
        self.connection.connect((self.host, self.port))

    def _receive(self):
        buffer = self.connection.recv(1024)
        r = json.loads(buffer)
        return r

    def send(self, method, params):
        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": int(time.time()*1000),
                "method": method,
                "params": params
            }
        ) + '\n'
        payload = payload.encode()
        self.connection.send(payload)
        return self._receive()


from os import path
from setuptools import (
    find_packages,
    setup,
)

# managing version
with open("VERSION", 'r') as f:
    VERSION = f.read()

# managing readme
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), 'r') as f:
    long_description = f.read()

setup(
    name='tbtc',
    version=VERSION,
    py_modules=['tbtc'],
    description="""A python library to interact with the tbtc protocol.""",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ankit Chiplunkar',
    author_email='ankitchiplunkar@gmail.com',
    url='https://github.com/ankitchiplunkar/tbtc.py',
    include_package_data=True,
    python_requires='>=3.6, <4',
    keywords=['ethereum', 'bitcoin'],
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "web3==5.12.1",
        "riemann-tx==2.1.0",
        "pytest==6.0.2",
        "pytest-cov==2.8.1",
      ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)



from tbtc.utils import initialize_contract
from tbtc.bitcoin_helpers import point_to_p2wpkh_address

class Deposit():
    """
    Deposit class to interact with the deposit contract
    """

    def __init__(self, tbtc_system, deposit_contract, keep_address):
        self.deposit_address = deposit_contract
        self.keep_address = keep_address
        self.tbtc_system = tbtc_system
        self.deposit_contract = initialize_contract(tbtc_system.w3, deposit_contract, "Deposit")
        self.keep_contract = initialize_contract(tbtc_system.w3, keep_address, "BondedECDSAKeep")
        
    def __repr__(self):
        return f"Deposit contract at: {self.deposit_contract}"

    def get_signer_public_key(self):
        # finding qn existing public key event
        event = self.tbtc_system._get_existing_public_key_event(self.deposit_address)
        # no registered public key found
        if event == []:
            # check if key is ready
            transaction_filter = self.keep_contract.events.PublicKeyPublished.createFilter(
                fromBlock='earliest',
                toBlock='latest'
            )
            key_ready_event = transaction_filter.get_all_entries()
            if key_ready_event is None:
                logger.info("Retry again later key has not been published yet")
                return None
            else:
                call = self.deposit_contract.functions.retrieveSignerPubkey()
                receipt = self.tbtc_system._manage_transaction(
                    call, 
                    gas_limit=160000
                    )
                event = self.tbtc_system.system.events.RegisteredPubkey().processReceipt(receipt)
        return point_to_p2wpkh_address(
            event[0]['args']['_signingGroupPubkeyX'], 
            event[0]['args']['_signingGroupPubkeyY']
            )

    def get_lot_size(self):
        return self.deposit_contract.functions.lotSizeSatoshis().call()